@@ -4,7 +4,7 @@
  *
  * This software is released under the PostgreSQL Licence.
  *
- * Guillaume Lelarge, guillaume@lelarge.info, 2011-2018.
+ * Guillaume Lelarge, guillaume@lelarge.info, 2011-2021.
  *
  * pgstats/pgcsvstat.c
  */
@@ -14,6 +14,9 @@
  * Headers
  */
 #include "postgres_fe.h"
+#include "common/string.h"
+
+#include <err.h>
 #include <sys/stat.h>
 
 #include <unistd.h>
@@ -28,7 +31,7 @@ extern char *optarg;
 /*
  * Defines
  */
-#define PGCSVSTAT_VERSION "1.0.1"
+#define PGCSVSTAT_VERSION "1.2.0"
 
 /* these are the opts structures for command line params */
 struct options
@@ -63,6 +66,10 @@ void		sql_exec_dump_pgstatbgwriter(void);
 void		sql_exec_dump_pgstatdatabase(void);
 void        sql_exec_dump_pgstatdatabaseconflicts(void);
 void        sql_exec_dump_pgstatreplication(void);
+void        sql_exec_dump_pgstatreplicationslots(void);
+void        sql_exec_dump_pgstatslru(void);
+void        sql_exec_dump_pgstatsubscription(void);
+void        sql_exec_dump_pgstatwal(void);
 void		sql_exec_dump_pgstatalltables(void);
 void		sql_exec_dump_pgstatallindexes(void);
 void		sql_exec_dump_pgstatioalltables(void);
@@ -77,6 +84,7 @@ bool		check_superuser(void);
 bool		backend_minimum_version(int major, int minor);
 bool        backend_has_pgstatstatements(void);
 
+
 /* function to parse command line options and check for some usage errors. */
 void
 get_opts(int argc, char **argv)
@@ -151,6 +159,7 @@ get_opts(int argc, char **argv)
 	}
 }
 
+
 static void
 help(const char *progname)
 {
@@ -205,6 +214,14 @@ sql_conn()
 	PGconn	   *my_conn;
 	char	   *password = NULL;
 	bool		new_pass;
+#if PG_VERSION_NUM >= 90300
+    const char **keywords;
+    const char **values;
+#else
+	int size;
+	char *dns;
+#endif
+	char		*message;
 
 	/*
 	 * Start the connection.  Loop until we have a password if requested by
@@ -212,31 +229,90 @@ sql_conn()
 	 */
 	do
 	{
-		new_pass = false;
-		my_conn = PQsetdbLogin(opts->hostname,
-							opts->port,
-							NULL,		/* options */
-							NULL,		/* tty */
-							opts->dbname,
-							opts->username,
-							password);
+
+#if PG_VERSION_NUM >= 90300
+		/*
+		 * We don't need to check if the database name is actually a complete
+		 * connection string, PQconnectdbParams being smart enough to check
+		 * this itself.
+		 */
+#define PARAMS_ARRAY_SIZE   8
+        keywords = pg_malloc(PARAMS_ARRAY_SIZE * sizeof(*keywords));
+        values = pg_malloc(PARAMS_ARRAY_SIZE * sizeof(*values));
+
+        keywords[0] = "host";
+        values[0] = opts->hostname,
+        keywords[1] = "port";
+        values[1] = opts->port;
+        keywords[2] = "user";
+        values[2] = opts->username;
+        keywords[3] = "password";
+        values[3] = password;
+        keywords[4] = "dbname";
+        values[4] = opts->dbname;
+        keywords[5] = "fallback_application_name";
+        values[5] = "pgcsvstat";
+        keywords[7] = NULL;
+        values[7] = NULL;
+
+        my_conn = PQconnectdbParams(keywords, values, true);
+#else
+		/* 34 is the length of the fallback application name setting */
+		size = 34;
+		if (opts->hostname)
+			size += strlen(opts->hostname) + 6;
+		if (opts->port)
+			size += strlen(opts->port) + 6;
+		if (opts->username)
+			size += strlen(opts->username) + 6;
+		if (opts->dbname)
+			size += strlen(opts->dbname) + 8;
+		dns = pg_malloc(size);
+		/*
+		 * Checking the presence of a = sign is our way to check that the
+		 * database name is actually a connection string. In such a case, we
+		 * keep this string as the connection string, and add other parameters
+		 * if they are supplied.
+		 */
+		sprintf(dns, "%s", "fallback_application_name='pgcsvstat' ");
+
+		if (strchr(opts->dbname, '=') != NULL)
+			sprintf(dns, "%s%s", dns, opts->dbname);
+		else if (opts->dbname)
+			sprintf(dns, "%sdbname=%s ", dns, opts->dbname);
+
+		if (opts->hostname)
+			sprintf(dns, "%shost=%s ", dns, opts->hostname);
+		if (opts->port)
+			sprintf(dns, "%sport=%s ", dns, opts->port);
+		if (opts->username)
+			sprintf(dns, "%suser=%s ", dns, opts->username);
+
+		if (opts->verbose)
+			printf("Connection string: %s\n", dns);
+
+		my_conn = PQconnectdb(dns);
+#endif
+
+        new_pass = false;
+
 		if (!my_conn)
 		{
-			fprintf(stderr, "%s: could not connect to database %s\n",
-					"pgstats", opts->dbname);
-			exit(1);
+			errx(1, "could not connect to database %s\n", opts->dbname);
 		}
 
 #if PG_VERSION_NUM >= 80200
 		if (PQstatus(my_conn) == CONNECTION_BAD &&
 			PQconnectionNeedsPassword(my_conn) &&
-			password == NULL)
+			!password)
 		{
 			PQfinish(my_conn);
 #if PG_VERSION_NUM < 100000
 			password = simple_prompt("Password: ", 100, false);
-#else
+#elif PG_VERSION_NUM < 140000
 			simple_prompt("Password: ", password, 100, false);
+#else
+			password = simple_prompt("Password: ", false);
 #endif
 			new_pass = true;
 		}
@@ -249,10 +325,9 @@ sql_conn()
 	/* check to see that the backend connection was successfully made */
 	if (PQstatus(my_conn) == CONNECTION_BAD)
 	{
-		fprintf(stderr, "%s: could not connect to database %s: %s",
-				"pgstats", opts->dbname, PQerrorMessage(my_conn));
+		message = PQerrorMessage(my_conn);
 		PQfinish(my_conn);
-		exit(1);
+		errx(1, "could not connect to database %s: %s", opts->dbname, message);
 	}
 
 	/* return the conn if good */
@@ -263,7 +338,7 @@ sql_conn()
  * Actual code to make call to the database and print the output data.
  */
 int
-sql_exec(const char *todo, const char* filename, bool quiet)
+sql_exec(const char *query, const char* filename, bool quiet)
 {
 	PGresult   *res;
 	FILE	   *fdcsv;
@@ -279,8 +354,8 @@ sql_exec(const char *todo, const char* filename, bool quiet)
 	fdcsv = fopen(filename, "a");
     if (!fdcsv)
     {
-        fprintf(stderr, "pgstats: fopen failed: %d\n", errno);
-        fprintf(stderr, "pgstats: filename was: %s\n", filename);
+        fprintf(stderr, "pgcsvstat: fopen failed: %d\n", errno);
+        fprintf(stderr, "pgcsvstat: filename was: %s\n", filename);
         
         PQfinish(conn);
         exit(-1);
@@ -291,13 +366,13 @@ sql_exec(const char *todo, const char* filename, bool quiet)
 	size = st.st_size;
 
 	/* make the call */
-	res = PQexec(conn, todo);
+	res = PQexec(conn, query);
 
 	/* check and deal with errors */
 	if (!res || PQresultStatus(res) > 2)
 	{
-		fprintf(stderr, "pgstats: query failed: %s\n", PQerrorMessage(conn));
-		fprintf(stderr, "pgstats: query was: %s\n", todo);
+		fprintf(stderr, "pgcsvstat: query failed: %s\n", PQerrorMessage(conn));
+		fprintf(stderr, "pgcsvstat: query was: %s\n", query);
 
 		PQclear(res);
 		PQfinish(conn);
@@ -347,33 +422,36 @@ sql_exec(const char *todo, const char* filename, bool quiet)
 void
 sql_exec_dump_pgstatactivity()
 {
-	char		todo[1024];
+	char		query[1024];
 	char		filename[1024];
 
 	/* get the oid and database name from the system pg_database table */
-	snprintf(todo, sizeof(todo),
-			 "SELECT date_trunc('seconds', now()), datid, datname, %s, "
+	snprintf(query, sizeof(query),
+			 "SELECT date_trunc('seconds', now()), datid, datname, %s, %s"
              "usesysid, usename, %s%s%s%s%s"
 			 "date_trunc('seconds', query_start) AS query_start, "
-             "%s%s%s%s%s "
+             "%s%s%s%s%s%s%s state "
              "FROM pg_stat_activity "
 			 "ORDER BY %s",
 		backend_minimum_version(9, 2) ? "pid" : "procpid",
+		backend_minimum_version(13, 0) ? "leader_pid, " : "",
 		backend_minimum_version(9, 0) ? "application_name, " : "",
 		backend_minimum_version(8, 1) ? "client_addr, " : "",
 		backend_minimum_version(9, 1) ? "client_hostname, " : "",
 		backend_minimum_version(8, 1) ? "client_port, date_trunc('seconds', backend_start) AS backend_start, " : "",
         backend_minimum_version(8, 3) ? "date_trunc('seconds', xact_start) AS xact_start, " : "",
         backend_minimum_version(9, 2) ? "state_change, " : "",
-        backend_minimum_version(8, 2) ? "waiting, " : "",
+        backend_minimum_version(9, 6) ? "wait_event_type, wait_event, " : backend_minimum_version(8, 2) ? "waiting, " : "",
         backend_minimum_version(9, 4) ? "backend_xid, " : "",
         backend_minimum_version(9, 4) ? "backend_xmin, " : "",
-        backend_minimum_version(9, 2) ? "query" : "current_query",
-		backend_minimum_version(9, 2) ? "pid" : "procpid");
+        backend_minimum_version(14, 0) ? "query_id, " : "",
+        backend_minimum_version(9, 2) ? "query, " : "current_query,",
+		backend_minimum_version(10, 0) ? "backend_type, " : "",
+		backend_minimum_version(9, 2) ? "pid" : "procpid");   // the last one is for the ORDER BY
 	snprintf(filename, sizeof(filename),
 			 "%s/pg_stat_activity.csv", opts->directory);
 
-	sql_exec(todo, filename, opts->quiet);
+	sql_exec(query, filename, opts->quiet);
 }
 
 /*
@@ -382,11 +460,11 @@ sql_exec_dump_pgstatactivity()
 void
 sql_exec_dump_pgstatbgwriter()
 {
-	char		todo[1024];
+	char		query[1024];
 	char		filename[1024];
 
 	/* get the oid and database name from the system pg_database table */
-	snprintf(todo, sizeof(todo),
+	snprintf(query, sizeof(query),
 			 "SELECT date_trunc('seconds', now()), checkpoints_timed, "
              "checkpoints_req, %sbuffers_checkpoint, buffers_clean, "
 			 "maxwritten_clean, buffers_backend, %sbuffers_alloc%s "
@@ -397,7 +475,7 @@ sql_exec_dump_pgstatbgwriter()
 	snprintf(filename, sizeof(filename),
 			 "%s/pg_stat_bgwriter.csv", opts->directory);
 
-	sql_exec(todo, filename, opts->quiet);
+	sql_exec(query, filename, opts->quiet);
 }
 
 /*
@@ -406,11 +484,11 @@ sql_exec_dump_pgstatbgwriter()
 void
 sql_exec_dump_pgstatarchiver()
 {
-	char		todo[1024];
+	char		query[1024];
 	char		filename[1024];
 
 	/* get the oid and database name from the system pg_database table */
-	snprintf(todo, sizeof(todo),
+	snprintf(query, sizeof(query),
 			 "SELECT date_trunc('seconds', now()), archived_count, "
              "last_archived_wal, date_trunc('seconds', last_archived_time) AS last_archived_time, "
 			 "failed_count, "
@@ -420,32 +498,35 @@ sql_exec_dump_pgstatarchiver()
 	snprintf(filename, sizeof(filename),
 			 "%s/pg_stat_archiver.csv", opts->directory);
 
-	sql_exec(todo, filename, opts->quiet);
+	sql_exec(query, filename, opts->quiet);
 }
 
 /*
  * Dump all databases stats.
+ * to be fixed wrt v14
  */
 void
 sql_exec_dump_pgstatdatabase()
 {
-	char		todo[1024];
+	char		query[1024];
 	char		filename[1024];
 
 	/* get the oid and database name from the system pg_database table */
-	snprintf(todo, sizeof(todo),
+	snprintf(query, sizeof(query),
 			 "SELECT date_trunc('seconds', now()), datid, datname, "
              "numbackends, xact_commit, xact_rollback, blks_read, blks_hit"
-             "%s%s%s "
+             "%s%s%s%s%s "
              "FROM pg_stat_database "
              "ORDER BY datname",
 		backend_minimum_version(8, 3) ? ", tup_returned, tup_fetched, tup_inserted, tup_updated, tup_deleted" : "",
 		backend_minimum_version(9, 1) ? ", conflicts, date_trunc('seconds', stats_reset) AS stats_reset" : "",
-		backend_minimum_version(9, 2) ? ", temp_files, temp_bytes, deadlocks, blk_read_time, blk_write_time" : "");
+		backend_minimum_version(9, 2) ? ", temp_files, temp_bytes, deadlocks, blk_read_time, blk_write_time" : "",
+		backend_minimum_version(12, 0) ? ", checksum_failures, checksum_last_failure" : "",
+		backend_minimum_version(14, 0) ? ", session_time, active_time, idle_in_transaction_time, sessions, sessions_abandoned, sessions_fatal, sessions_killed" : "");
 	snprintf(filename, sizeof(filename),
 			 "%s/pg_stat_database.csv", opts->directory);
 
-	sql_exec(todo, filename, opts->quiet);
+	sql_exec(query, filename, opts->quiet);
 }
 
 /*
@@ -454,18 +535,18 @@ sql_exec_dump_pgstatdatabase()
 void
 sql_exec_dump_pgstatdatabaseconflicts()
 {
-	char		todo[1024];
+	char		query[1024];
 	char		filename[1024];
 
 	/* get the oid and database name from the system pg_database table */
-	snprintf(todo, sizeof(todo),
+	snprintf(query, sizeof(query),
 			 "SELECT date_trunc('seconds', now()), * "
              "FROM pg_stat_database_conflicts "
              "ORDER BY datname");
 	snprintf(filename, sizeof(filename),
 			 "%s/pg_stat_database_conflicts.csv", opts->directory);
 
-	sql_exec(todo, filename, opts->quiet);
+	sql_exec(query, filename, opts->quiet);
 }
 
 /*
@@ -474,26 +555,123 @@ sql_exec_dump_pgstatdatabaseconflicts()
 void
 sql_exec_dump_pgstatreplication()
 {
-	char		todo[1024];
+	char		query[1024];
 	char		filename[1024];
 
 	/* get the oid and database name from the system pg_database table */
-	snprintf(todo, sizeof(todo),
+	snprintf(query, sizeof(query),
 			 "SELECT date_trunc('seconds', now()), %s, usesysid, usename, "
              "application_name, client_addr, client_hostname, client_port, "
              "date_trunc('seconds', backend_start) AS backend_start, %sstate, "
-             "pg_current_xlog_location() AS master_location, "
-             "sent_location, write_location, flush_location, replay_location, "
+             "%s AS master_location, %s%s"
              "sync_priority, "
-             "sync_state "
+             "sync_state%s "
              "FROM pg_stat_replication "
              "ORDER BY application_name",
 		backend_minimum_version(9, 2) ? "pid" : "procpid",
-		backend_minimum_version(9, 4) ? "backend_xmin, " : "");
+		backend_minimum_version(9, 4) ? "backend_xmin, " : "",
+		backend_minimum_version(10, 0) ? "pg_current_wal_lsn()" : "pg_current_xlog_location()",
+		backend_minimum_version(10, 0) ? "sent_lsn, write_lsn, flush_lsn, replay_lsn, " :
+		     "sent_location, write_location, flush_location, replay_location, ",
+		backend_minimum_version(10, 0) ? "write_lag, flush_lag, replay_lag, " : "",
+		backend_minimum_version(12, 0) ? ", reply_time" : "");
 	snprintf(filename, sizeof(filename),
 			 "%s/pg_stat_replication.csv", opts->directory);
 
-	sql_exec(todo, filename, opts->quiet);
+	sql_exec(query, filename, opts->quiet);
+}
+
+/*
+ * Dump all replication slots stats.
+ */
+void
+sql_exec_dump_pgstatreplicationslots()
+{
+	char		query[1024];
+	char		filename[1024];
+
+	/* get the oid and database name from the system pg_database table */
+	snprintf(query, sizeof(query),
+			 "SELECT date_trunc('seconds', now()), slot_name, "
+			 "spill_txns, spill_count, spill_bytes, "
+			 "stream_txns, stream_count, stream_bytes, "
+			 "total_txns, total_bytes, "
+			 "date_trunc('seconds', stats_reset) AS stats_reset "
+             "FROM pg_stat_replication_slots "
+             "ORDER BY slot_name");
+	snprintf(filename, sizeof(filename),
+			 "%s/pg_stat_replication_slots.csv", opts->directory);
+
+	sql_exec(query, filename, opts->quiet);
+}
+
+/*
+ * Dump all SLRU stats.
+ */
+void
+sql_exec_dump_pgstatslru()
+{
+	char		query[1024];
+	char		filename[1024];
+
+	/* get the oid and database name from the system pg_database table */
+	snprintf(query, sizeof(query),
+			 "SELECT date_trunc('seconds', now()), name, "
+			 "blks_zeroed, blks_hit, blks_read, blks_written, blks_exists, "
+			 "flushes, truncates, "
+			 "date_trunc('seconds', stats_reset) AS stats_reset "
+             "FROM pg_stat_slru "
+             "ORDER BY name");
+	snprintf(filename, sizeof(filename),
+			 "%s/pg_stat_slru.csv", opts->directory);
+
+	sql_exec(query, filename, opts->quiet);
+}
+
+/*
+ * Dump all subscriptions stats.
+ */
+void
+sql_exec_dump_pgstatsubscription()
+{
+	char		query[1024];
+	char		filename[1024];
+
+	/* get the oid and database name from the system pg_database table */
+	snprintf(query, sizeof(query),
+			 "SELECT date_trunc('seconds', now()), subid, subname, "
+			 "pid, relid, received_lsn, "
+			 "date_trunc('seconds', last_msg_send_time) AS last_msg_send_time, "
+			 "date_trunc('seconds', last_msg_receipt_time) AS last_msg_receipt_time, "
+			 "latest_end_lsn, date_trunc('seconds', latest_end_time) AS latest_end_time "
+             "FROM pg_stat_subscription "
+             "ORDER BY subid");
+	snprintf(filename, sizeof(filename),
+			 "%s/pg_stat_subscription.csv", opts->directory);
+
+	sql_exec(query, filename, opts->quiet);
+}
+
+/*
+ * Dump all WAL stats.
+ */
+void
+sql_exec_dump_pgstatwal()
+{
+	char		query[1024];
+	char		filename[1024];
+
+	/* get the oid and database name from the system pg_database table */
+	snprintf(query, sizeof(query),
+			 "SELECT date_trunc('seconds', now()), "
+			 "wal_records, wal_fpi, wal_bytes, wal_buffers_full, wal_write, "
+			 "wal_sync, wal_write_time, wal_sync_time, "
+			 "date_trunc('seconds', stats_reset) AS stats_reset "
+             "FROM pg_stat_wal");
+	snprintf(filename, sizeof(filename),
+			 "%s/pg_stat_wal.csv", opts->directory);
+
+	sql_exec(query, filename, opts->quiet);
 }
 
 /*
@@ -502,11 +680,11 @@ sql_exec_dump_pgstatreplication()
 void
 sql_exec_dump_pgstatalltables()
 {
-	char		todo[1024];
+	char		query[1024];
 	char		filename[1024];
 
 	/* get the oid and database name from the system pg_database table */
-	snprintf(todo, sizeof(todo),
+	snprintf(query, sizeof(query),
 			 "SELECT date_trunc('seconds', now()), relid, schemaname, relname, "
              "seq_scan, seq_tup_read, idx_scan, idx_tup_fetch, n_tup_ins, "
              "n_tup_upd, n_tup_del"
@@ -514,17 +692,19 @@ sql_exec_dump_pgstatalltables()
 			 "%s"
 			 "%s"
 			 "%s"
+			 "%s"
              " FROM pg_stat_all_tables "
 	     "WHERE schemaname <> 'information_schema' "
 	     "ORDER BY schemaname, relname",
 		backend_minimum_version(8, 3) ? ", n_tup_hot_upd, n_live_tup, n_dead_tup" : "",
         backend_minimum_version(9, 4) ? ", n_mod_since_analyze" : "",
+        backend_minimum_version(13, 0) ? ", n_ins_since_vacuum" : "",
         backend_minimum_version(8, 2) ? ", date_trunc('seconds', last_vacuum) AS last_vacuum, date_trunc('seconds', last_autovacuum) AS last_autovacuum, date_trunc('seconds',last_analyze) AS last_analyze, date_trunc('seconds',last_autoanalyze) AS last_autoanalyze" : "",
         backend_minimum_version(9, 1) ? ", vacuum_count, autovacuum_count, analyze_count, autoanalyze_count" : "");
 	snprintf(filename, sizeof(filename),
 			 "%s/pg_stat_all_tables.csv", opts->directory);
 
-	sql_exec(todo, filename, opts->quiet);
+	sql_exec(query, filename, opts->quiet);
 }
 
 /*
@@ -533,11 +713,11 @@ sql_exec_dump_pgstatalltables()
 void
 sql_exec_dump_pgstatallindexes()
 {
-	char		todo[1024];
+	char		query[1024];
 	char		filename[1024];
 
 	/* get the oid and database name from the system pg_database table */
-	snprintf(todo, sizeof(todo),
+	snprintf(query, sizeof(query),
 			 "SELECT date_trunc('seconds', now()), * "
 			 "FROM pg_stat_all_indexes "
 	     		 "WHERE schemaname <> 'information_schema' "
@@ -545,7 +725,7 @@ sql_exec_dump_pgstatallindexes()
 	snprintf(filename, sizeof(filename),
 			 "%s/pg_stat_all_indexes.csv", opts->directory);
 
-	sql_exec(todo, filename, opts->quiet);
+	sql_exec(query, filename, opts->quiet);
 }
 
 /*
@@ -554,11 +734,11 @@ sql_exec_dump_pgstatallindexes()
 void
 sql_exec_dump_pgstatioalltables()
 {
-	char		todo[1024];
+	char		query[1024];
 	char		filename[1024];
 
 	/* get the oid and database name from the system pg_database table */
-	snprintf(todo, sizeof(todo),
+	snprintf(query, sizeof(query),
 			 "SELECT date_trunc('seconds', now()), * "
 			 "FROM pg_statio_all_tables "
 	     		 "WHERE schemaname <> 'information_schema' "
@@ -566,7 +746,7 @@ sql_exec_dump_pgstatioalltables()
 	snprintf(filename, sizeof(filename),
 			 "%s/pg_statio_all_tables.csv", opts->directory);
 
-	sql_exec(todo, filename, opts->quiet);
+	sql_exec(query, filename, opts->quiet);
 }
 
 /*
@@ -575,11 +755,11 @@ sql_exec_dump_pgstatioalltables()
 void
 sql_exec_dump_pgstatioallindexes()
 {
-	char		todo[1024];
+	char		query[1024];
 	char		filename[1024];
 
 	/* get the oid and database name from the system pg_database table */
-	snprintf(todo, sizeof(todo),
+	snprintf(query, sizeof(query),
 			 "SELECT date_trunc('seconds', now()), * "
 			 "FROM pg_statio_all_indexes "
 	     		 "WHERE schemaname <> 'information_schema' "
@@ -587,7 +767,7 @@ sql_exec_dump_pgstatioallindexes()
 	snprintf(filename, sizeof(filename),
 			 "%s/pg_statio_all_indexes.csv", opts->directory);
 
-	sql_exec(todo, filename, opts->quiet);
+	sql_exec(query, filename, opts->quiet);
 }
 
 /*
@@ -596,11 +776,11 @@ sql_exec_dump_pgstatioallindexes()
 void
 sql_exec_dump_pgstatioallsequences()
 {
-	char		todo[1024];
+	char		query[1024];
 	char		filename[1024];
 
 	/* get the oid and database name from the system pg_database table */
-	snprintf(todo, sizeof(todo),
+	snprintf(query, sizeof(query),
 			 "SELECT date_trunc('seconds', now()), * "
 			 "FROM pg_statio_all_sequences "
 	     		 "WHERE schemaname <> 'information_schema' "
@@ -608,7 +788,7 @@ sql_exec_dump_pgstatioallsequences()
 	snprintf(filename, sizeof(filename),
 			 "%s/pg_statio_all_sequences.csv", opts->directory);
 
-	sql_exec(todo, filename, opts->quiet);
+	sql_exec(query, filename, opts->quiet);
 }
 
 /*
@@ -617,11 +797,11 @@ sql_exec_dump_pgstatioallsequences()
 void
 sql_exec_dump_pgstatuserfunctions()
 {
-	char		todo[1024];
+	char		query[1024];
 	char		filename[1024];
 
 	/* get the oid and database name from the system pg_database table */
-	snprintf(todo, sizeof(todo),
+	snprintf(query, sizeof(query),
 			 "SELECT date_trunc('seconds', now()), * "
 			 "FROM pg_stat_user_functions "
 	     		 "WHERE schemaname <> 'information_schema' "
@@ -629,7 +809,7 @@ sql_exec_dump_pgstatuserfunctions()
 	snprintf(filename, sizeof(filename),
 			 "%s/pg_stat_user_functions.csv", opts->directory);
 
-	sql_exec(todo, filename, opts->quiet);
+	sql_exec(query, filename, opts->quiet);
 }
 
 /*
@@ -638,11 +818,11 @@ sql_exec_dump_pgstatuserfunctions()
 void
 sql_exec_dump_pgclass_size()
 {
-	char		todo[1024];
+	char		query[1024];
 	char		filename[1024];
 
 	/* get the oid and database name from the system pg_database table */
-	snprintf(todo, sizeof(todo),
+	snprintf(query, sizeof(query),
 			 "SELECT date_trunc('seconds', now()), n.nspname, c.relname, c.relkind, c.reltuples, c.relpages%s "
 			 "FROM pg_class c, pg_namespace n "
 			 "WHERE n.oid=c.relnamespace AND n.nspname <> 'information_schema' "
@@ -651,32 +831,37 @@ sql_exec_dump_pgclass_size()
 	snprintf(filename, sizeof(filename),
 			 "%s/pg_class_size.csv", opts->directory);
 
-	sql_exec(todo, filename, opts->quiet);
+	sql_exec(query, filename, opts->quiet);
 }
 
 /*
  * Dump all statements stats.
+ * to be fixed wrt v14
  */
 void
 sql_exec_dump_pgstatstatements()
 {
-	char		todo[1024];
+	char		query[1024];
 	char		filename[1024];
 
 	/* get the oid and database name from the system pg_database table */
-	snprintf(todo, sizeof(todo),
+	snprintf(query, sizeof(query),
 			 "SELECT date_trunc('seconds', now()), r.rolname, d.datname, "
-             "regexp_replace(query, E'\n', ' ', 'g') as query, calls, total_time, rows, "
+             "%sregexp_replace(query, E'\n', ' ', 'g') as query, %scalls, %s, rows, "
              "shared_blks_hit, shared_blks_read, shared_blks_written, "
              "local_blks_hit, local_blks_read, local_blks_written, "
-             "temp_blks_read, temp_blks_written "
+             "temp_blks_read, temp_blks_written%s "
 			 "FROM pg_stat_statements q, pg_database d, pg_roles r "
 			 "WHERE q.userid=r.oid and q.dbid=d.oid "
-			 "ORDER BY r.rolname, d.datname");
+			 "ORDER BY r.rolname, d.datname",
+		backend_minimum_version(14, 0) ? "toplevel, queryid, " : "",
+		backend_minimum_version(13, 0) ? "plans, total_plan_time, min_plan_time, max_plan_time, mean_plan_time, stddev_plan_time, " : "",
+		backend_minimum_version(13, 0) ? "total_exec_time, min_exec_time, max_exec_time, mean_exec_time, stddev_exec_time" : "total_time",
+		backend_minimum_version(14, 0) ? ", wal_records, wal_fpi, wal_bytes" : "");
 	snprintf(filename, sizeof(filename),
 			 "%s/pg_stat_statements.csv", opts->directory);
 
-	sql_exec(todo, filename, opts->quiet);
+	sql_exec(query, filename, opts->quiet);
 }
 
 /*
@@ -685,11 +870,19 @@ sql_exec_dump_pgstatstatements()
 void
 sql_exec_dump_xlog_stat()
 {
-	char		todo[1024];
+	char		query[1024];
 	char		filename[1024];
 
 	/* get the oid and database name from the system pg_database table */
-	snprintf(todo, sizeof(todo),
+	snprintf(query, sizeof(query),
+		backend_minimum_version(10, 0)
+		?
+			 "SELECT date_trunc('seconds', now()), pg_walfile_name(pg_current_wal_lsn())=pg_ls_dir AS current, pg_ls_dir AS filename, "
+			 "(SELECT modification FROM pg_stat_file('pg_wal/'||pg_ls_dir)) AS modification_timestamp "
+			 "FROM pg_ls_dir('pg_wal') "
+			 "WHERE pg_ls_dir ~ E'^[0-9A-F]{24}' "
+			 "ORDER BY pg_ls_dir"
+		:
 			 "SELECT date_trunc('seconds', now()), pg_xlogfile_name(pg_current_xlog_location())=pg_ls_dir AS current, pg_ls_dir AS filename, "
 			 "(SELECT modification FROM pg_stat_file('pg_xlog/'||pg_ls_dir)) AS modification_timestamp "
 			 "FROM pg_ls_dir('pg_xlog') "
@@ -698,7 +891,7 @@ sql_exec_dump_xlog_stat()
 	snprintf(filename, sizeof(filename),
 			 "%s/pg_xlog_stat.csv", opts->directory);
 
-	sql_exec(todo, filename, opts->quiet);
+	sql_exec(query, filename, opts->quiet);
 }
 
 
@@ -708,20 +901,20 @@ sql_exec_dump_xlog_stat()
 void
 fetch_version()
 {
-	char		todo[1024];
+	char		query[1024];
 	PGresult   *res;
 
 	/* get the oid and database name from the system pg_database table */
-	snprintf(todo, sizeof(todo), "SELECT version()");
+	snprintf(query, sizeof(query), "SELECT version()");
 
 	/* make the call */
-	res = PQexec(conn, todo);
+	res = PQexec(conn, query);
 
 	/* check and deal with errors */
 	if (!res || PQresultStatus(res) > 2)
 	{
-		fprintf(stderr, "pgstats: query failed: %s\n", PQerrorMessage(conn));
-		fprintf(stderr, "pgstats: query was: %s\n", todo);
+		fprintf(stderr, "pgcsvstat: query failed: %s\n", PQerrorMessage(conn));
+		fprintf(stderr, "pgcsvstat: query was: %s\n", query);
 
 		PQclear(res);
 		PQfinish(conn);
@@ -760,8 +953,8 @@ check_superuser()
 	/* check and deal with errors */
 	if (!res || PQresultStatus(res) > 2)
 	{
-		fprintf(stderr, "pgstats: query failed: %s\n", PQerrorMessage(conn));
-		fprintf(stderr, "pgstats: query was: %s\n", sql);
+		fprintf(stderr, "pgcsvstat: query failed: %s\n", PQerrorMessage(conn));
+		fprintf(stderr, "pgcsvstat: query was: %s\n", sql);
 
 		PQclear(res);
 		PQfinish(conn);
@@ -811,8 +1004,8 @@ backend_has_pgstatstatements()
 	/* check and deal with errors */
 	if (!res || PQresultStatus(res) > 2)
 	{
-		fprintf(stderr, "pgstats: query failed: %s\n", PQerrorMessage(conn));
-		fprintf(stderr, "pgstats: query was: %s\n", sql);
+		fprintf(stderr, "pgcsvstat: query failed: %s\n", PQerrorMessage(conn));
+		fprintf(stderr, "pgcsvstat: query was: %s\n", sql);
 
 		PQclear(res);
 		PQfinish(conn);
@@ -859,8 +1052,10 @@ main(int argc, char **argv)
     /* check superuser attribute */
     is_superuser = check_superuser();
 
-	/* grabe cluster stats info */
+	/* grab cluster stats info */
 	sql_exec_dump_pgstatactivity();
+	if (backend_minimum_version(9, 4))
+		sql_exec_dump_pgstatarchiver();
 	if (backend_minimum_version(8, 3))
 		sql_exec_dump_pgstatbgwriter();
 	sql_exec_dump_pgstatdatabase();
@@ -869,10 +1064,17 @@ main(int argc, char **argv)
 		sql_exec_dump_pgstatdatabaseconflicts();
 		sql_exec_dump_pgstatreplication();
     }
-	if (backend_minimum_version(9, 4))
-		sql_exec_dump_pgstatarchiver();
+	if (backend_minimum_version(14, 0))
+		sql_exec_dump_pgstatreplicationslots();
+	if (backend_minimum_version(13, 0))
+		sql_exec_dump_pgstatslru();
+	if (backend_minimum_version(10, 0))
+		sql_exec_dump_pgstatsubscription();
+	if (backend_minimum_version(14, 0))
+		sql_exec_dump_pgstatwal();
+    /* TODO pg_stat_wal_receiver */
 
-	/* grabe database stats info */
+	/* grab database stats info */
 	sql_exec_dump_pgstatalltables();
 	sql_exec_dump_pgstatallindexes();
 	sql_exec_dump_pgstatioalltables();
@@ -881,7 +1083,10 @@ main(int argc, char **argv)
     if (backend_minimum_version(8, 4))
 		sql_exec_dump_pgstatuserfunctions();
 
-	/* grabe other informations */
+    /* grab progress stats info */
+    /* TODO */
+
+	/* grab other informations */
 	sql_exec_dump_pgclass_size();
     if (backend_has_pgstatstatements())
 	    sql_exec_dump_pgstatstatements();
